#!/usr/bin/env python3

# this filter demonstrates a (rather whacky and extreme) use of 'cloak_and_dagger'
# cloak_and_dagger is used for
#   - finding the working directories (local and remote)
#   - finding the legal ftp commands
#   - completing (local or remote) filenames and directories

"""
A vanilla ftp is not bundled in modern Linux today.
For example, ftp on Fedora 21 is readline-enabled.

$ ldd /usr/bin/ftp 
        linux-vdso.so.1 =>  (0x00007ffd015e7000)
        libreadline.so.6 => /lib64/libreadline.so.6 (0x00007fd213777000)
        libncurses.so.5 => /lib64/libncurses.so.5 (0x00007fd21354f000)
        libtinfo.so.5 => /lib64/libtinfo.so.5 (0x00007fd213324000)
        libc.so.6 => /lib64/libc.so.6 (0x00007fd212f67000)
        libdl.so.2 => /lib64/libdl.so.2 (0x00007fd212d63000)
        /lib64/ld-linux-x86-64.so.2 (0x00007fd2139dd000)

However, the feature can be disabled by '-e' option in Fedora case.
This example was tested on the pair of ftp bundled in RHEL6/7, Fedora21 and vsftpd in RHEL6.
Please see `man ftp` to find out how to disable it, if this filter does not work for you.
"""

import sys
import os

if 'RLWRAP_FILTERDIR' in os.environ:
    sys.path.append(os.environ['RLWRAP_FILTERDIR'])
else:
    sys.path.append('.')

import rlwrapfilter
import re


############################ subroutines ####################################

def listing (dir, where, what):
    dir = '.' if dir == None else dir

    command = '!ls -la {0}|cat'.format(dir) if where == 'local' else 'ls -la'
    lines = re.split(r'\r?\n', filter.cloak_and_dagger(command, ftp_prompt, 2))
    if dir_filename_column[where] == None: # find out which column of listing has the filename
        dotdotline = [l for l in lines if re.search(r'(^|\s+)\.\.(\s|$)', l)][0]; # .. should always be there 
        fields = re.split(r'\s+', dotdotline)
        try:
            dir_filename_column[where] = fields.index("..")
        except ValueError as e:
            rlwrapfilter.rlwrap_error(
                "couldn't determine filename column of $where listing", e)

    pattern = "^d" if what == "directories" else "^-"
    lines = [l for l in lines if re.search(pattern, l)] # makes only sense if there is a column with drwxr-xr-x
    results = [re.split(r'\s+', l)[dir_filename_column[where]] for l in lines]
    return results


def pwd(where):
    command = "!pwd" if where == "local" else "pwd"
    result = filter.cloak_and_dagger(command, ftp_prompt, 1)
    pattern = "(.*?)\r?\n" if where == "local" else '"(.*?)"'
    m = re.search(pattern, result)
    pwd = m.group(1)
    return pwd


def prompt_handler(prompt):
    if prompt == ftp_prompt:
        if len(ftp_commands) == 0:
            ftp_commands.extend(commands())
    else:
        return prompt
    local = pwd('local')
    remote = pwd('remote')
    local = local.replace(os.environ['HOME'], '~')
    rtext = '(remote: ' + remote + ')' if remote else '(not connected)'
    return '{0} {1}> '.format(local,rtext)


"""
sub test {
  my $listing = join ', ', listing (".", "remote", "directories");
  $filter -> send_output_oob("\n Hier: <$listing>\n");
}

"""

def commands():
    help_text = filter.cloak_and_dagger("help", ftp_prompt, 0.5)
    lines = help_text.split('\n')
    del lines[0:2] # remove the first 2 lines
    commands = [command for command in re.split(r'\s+', ''.join(lines))]
    return commands

def complete_handler(line, prefix, completions):

    nwords = len(re.split(r'\s+', line))
    if prefix == None:
        nwords += 1 # TAB at start of a new (empty) argument

    if nwords <= 1:
        completions.extend([c for c in ftp_commands if re.search(r'^' + prefix, c)])
        return completions

    command = re.search(r'\s*(\S+)', line).group(1)
    prefix_match = re.search(r'((.*)/)?([^/]*)', prefix)
    dir = prefix_match.group(2)
    name_prefix = prefix_match.group(3)
    #dir = '.' if dir == None else dir
    dir = '' if dir == None else dir
    narg = nwords-2
    try:
        completion_types[command][narg*2]
        candidates = listing(dir,
                             completion_types[command][2*narg],
                             completion_types[command][2*narg+1])
        completions.extend([os.path.join(dir, c) for c in candidates if re.search(r'^' + name_prefix, c)])
    except ValueError as e:
        pass
    completions = list(set(completions))
    return completions


if __name__ == '__main__':
    if __file__ == "./doctest":
        import doctest
        doctest.testmod()
        sys.exit(0)


    ftp_prompt = "ftp> "
    ftp_commands = []
    completion_types = {
        'cd'  : ['remote', 'directories'],
        'lcd' : ['local', 'directories'],
        'get' : ['remote','files','local','directories'],
        'put' : ['local', 'files','remote','directories']
    }
    
    
    filter = rlwrapfilter.RlwrapFilter()
    dir_filename_column = {'local':None, 'remote':None}
    #dir_filename_column['local'] = None
    #dir_filename_column['remote'] = None
    
    filter.help_text = '\n'.join([
        "usage: rlwrap -c [-aword:] -z ftp_filter.py ftp (-e) [hostname]",
        "run plain Netkit ftp with completion for commands, local and remote files",
        "(demo filter to show the use of the cloak_and_dagger method)"])
    filter.prompt_handler = prompt_handler
    filter.completion_handler = complete_handler
    filter.cloak_and_dagger_verbose = False # set to True to spy on cloak_and_dagger dialogue
    
    
    if (not 'RLWRAP_COMMAND_PID' in os.environ) or re.match(r'^ftp', os.environ['RLWRAP_COMMAND_LINE']):
        raise SystemExit("This filter works only with plain vanilla ftp\n")

    filter.run()



#!/usr/bin/env python3

"""
a demo for painting prompt

this filter tests rlwrap's tolerance for very heavily cooked prompts.
it uses the seldom-used xterm 256-colour mode (which may or may not work in your 
other terminal emulators like rxvt) 

try `rlwrap -z './paint_prompt.py 0087ff--ff0000' telnet`
"""

import sys
import os

if 'RLWRAP_FILTERDIR' in os.environ:
    sys.path.append(os.environ['RLWRAP_FILTERDIR'])
else:
    sys.path.append('.')

import rlwrapfilter
import re

rampsize = 0xe7 - 0x10; # works on my gnome terminal

# create table to map rgb to xterm color id, eg,
# rgb2xterm['000000'] == '16', rgb2xterm['00005f'] == '17',...etc
rgb2xterm = {}
xtermcolor = 16
for r in (0, 0x5f, 0x87, 0xaf, 0xdf, 0xff):
    for g in (0, 0x5f, 0x87, 0xaf, 0xdf, 0xff):
        for b in (0, 0x5f, 0x87, 0xaf, 0xdf, 0xff):
            rgb = hex((r<<16) + (g<<8) + b)[2:] # remove the prefix '0x'
            rgb2xterm[rgb] = str(xtermcolor)
            xtermcolor += 1

filter = rlwrapfilter.RlwrapFilter()
name = filter.name
filter.help_text = '\n'.join([
    "Usage: rlwrap -z '$name <colour1>--<colour2> <command>",
    "paint the prompt in colour gradient between colours <colour1> and <colour2>",
    "colour names must be a 6 digit hex code of rgb,",
    "e.g. 00ff00--ff0000"
    ])

m = re.match(r'(.*)--(.*)', sys.argv[1])
colour1 = m.group(1)
colour2 = m.group(2)


def ramp(val1, val2, frac):
    """intermediate value 100*$frac % between $val1 and $val2"""
    val = int(val1 + frac * (val2-val1))

    # round off
    if (0 <= val < (0+0x5f)/2):
        val = 0x0
    elif ((0+0x5f)/2<= val < (0x5f+0x87)/2):
        val = 0x5f
    elif ((0x5f+0x87)/2<= val < (0x87+0xaf)/2):
        val = 0x87
    elif ((0x87+0xaf)/2<= val < (0xaf+0xdf)/2):
        val = 0xaf
    elif ((0xaf+0xdf)/2<= val < (0xdf+0xff)/2):
        val = 0xdf
    elif ((0xdf+0xff)/2<= val <= 0xff):
        val = 0xff
    return val


def intermediate_color(frac):
    r1 = int(colour1[0:2], 16)
    g1 = int(colour1[2:4], 16)
    b1 = int(colour1[4:6], 16)
    r2 = int(colour2[0:2], 16)
    g2 = int(colour2[2:4], 16)
    b2 = int(colour2[4:6], 16)
    r = ramp(r1, r2, frac)
    g = ramp(g1, g2, frac)
    b = ramp(b1, b2, frac)
    rgb = (r<<16) + (g<<8) + b
    return hex(rgb)[2:] # remove the prefix '0x'


def apply_ramp(text):
    buf = []

    # text -> ['t', 'e', 'x', 't']
    i = 0
    for c in text:
        # ESC[38;5;$colour is the xterm code to swithc to $colour, the \x01 and \x02 are
        # readline "ignore codes" to tell readline that the sequence is unprintable 
        buf.append(
            ''.join([
                "\x01\x1b[38;5;{0}m\x02".format(rgb2xterm[intermediate_color(i/len(text))]),
                c
            ]))
        i += 1

    return ''.join(buf) + "\x01\x1b[0m\x02"


filter.prompt_handler = apply_ramp

filter.run()



#!/usr/bin/env python3

"""handle hotkeys"""

import subprocess
import sys
import os
import tempfile
import re

if 'RLWRAP_FILTERDIR' in os.environ:
    sys.path.append(os.environ['RLWRAP_FILTERDIR'])
else:
    sys.path.append('.')

import rlwrapfilter


# A hotkey handler is called with five parameters:
#   1: the key (e.g. 14 for CTRL+M)
#   2: the prefix, i.e. the input line up to (but not including) the cursor
#   3: the postfix: the rest of the input line (without a concluding newline, of course)
#   4: the whole history (all lines, oldest first, interspersed with newlines: "line 1\nline2\n ...line N")
#   5: the history position at the moment of the hotkey press (oldest line = 0)
#
# If the hotkey was bound to "rlwrap-hotkey-without-history" the last two parameters will be empty and can be ignored
# The return value is a similar list (where all values may be changed: the input line could be re-written,
# the history revised, etc. The first parameter makes no sense as a return value: if it is != $key
# it will pe put in the "echo area". If the key was bound to rlwrap-hotkey-without-history the last two elements of the 
# returned list are ignored.
#
# If the postfix is returned with a concluding newline, the resulting input line is accepted immediately, otherwise
# it is put in readlines input buffer again, with the cursor at the beginning of the returned postfix 
#
# Summary: ($echo, $prefix, $postfix, $history, $histpos) = handler($key, $prefix, $postfix, $history, $histpos)


# generic hotkey handler, that dispatches on the value of $key (using the hash %$keymap defined at the top of this file

@rlwrapfilter.intercept_error
def hotkey(key, *other_params): # key = e.g.  "<CTRL-Y>"
    ukey = uncontrol(key)
    if ukey not in keymap:
        return (key, *other_params) # a filter further downstream may want to handle this hotkey
    handler = keymap[ukey]
    result = handler(0, *other_params)
    return result


############################# A few handlers ###############################################
#
# After dispatch (on the value of $key) by the hotkey() function the value of $key is not relevant anymore.
# its place is now taken by a parameter $doc :
#
# ($echo, $prefix, $postfix, $history, $histpos) = myfunc(0,  $prefix, $postfix, $history, $histpos)
# "docstring"                                    = myfunc(1, @not_interesting)



def yank_clipboard(doc, prefix, postfix, *boring):
    if doc:
        return "insert from clipboard"
    selection = safe_backtick(qw('xsel -o'))
    return ("", prefix + selection, postfix, *boring)


def date_in_echo_area(doc, *boring):
    if doc:
        return "show current time in echo area"
    date = safe_backtick(qw('date +%H:%M'))
    return ("({0}) ".format(date), *boring)


def edit_history(doc, prefix, postfix, history, histpos):
    if doc:
        return "edit current history"
    if not histpos.isdigit():
        sys.exit("$histpos is not a number - did you bind this key to 'rlwrap-hotkey-without-history'?")
    editfile = tempfile.NamedTemporaryFile()
    editfilename = editfile.name
    lineno = int(histpos) + 1
    colno = len(prefix) + 1
    editfile.write(history.encode(sys.stdin.encoding))
    editfile.flush()
    if 'RLWRAP_EDITOR' in os.environ:
        editor = os.environ['RLWRAP_EDITOR']
    else:
        editor = "vi +%L"
    editor = editor.replace('%L', str(lineno))
    editor = editor.replace('%C', str(colno))
    os.system(editor + ' ' + editfilename)
    editfile.seek(0,0)
    lines = map(lambda l: l.decode(sys.stdin.encoding), editfile.readlines())
    new_history = []
    counter = 0
    empty_counter = 0
    last_counter = 0
    last_empty_counter = 0
    for line in lines:
        line = line.replace('\t', '')
        line = line.lstrip()
        line = line.rstrip()
        if not line == '':
            if (empty_counter > 0):
                # remember position of last line after an empty line,
                # and the number of empty lines:
                (last_counter, last_empty_counter) = (counter, empty_counter)
            empty_counter = 0;
            counter = counter + 1 # We count 0-based, so increment only now
            new_history.append(line)
        else:
            empty_counter = empty_counter + 1
    if last_empty_counter > 0:
        histpos = str(last_counter)
        prefix = new_history[last_counter]
        postfix = ""
    return ("", prefix, postfix, '\n'.join(new_history), histpos)




def split_off_last_word(string):
    '''split_off_last_word("In the gener") = ("In the ", "gener") '''
    break_chars = os.environ['RLWRAP_BREAK_CHARS'] if 'RLWRAP_BREAK_CHARS' in os.environ else " \t\n" # old rlwrap with newer filter - use a sensible default
    break_chars =  re.sub(r'(.)', r'\\1', break_chars)    # escape all chars in break_chars  
    break_chars = break_chars or ' ' # prevent python from choking on a bad regex '[]' in the next line
    words       = re.split('[{}]'.format(break_chars), string)
    last_word  = words[-1]
    return (string[0:len(string)-len(last_word)], last_word)


def fuzzy_filter_history(doc, prefix, postfix, history, histpos, command):
    '''filter history through command (either 'peco' or the very similar  'fzf') '''
    if doc:
        return "{} current history".format(command)
    first_chunk, last_word  = split_off_last_word(prefix)
    command_line = [command, '--select-1', '--query',  last_word ]
    select_1 = ''
    with subprocess.Popen(command_line
                          ,stdin=subprocess.PIPE
                          ,stdout=subprocess.PIPE
                          ,universal_newlines=True) as p:
        (select_1, error) = p.communicate(input=history)
    select_1 = select_1.rstrip()
    return ("", first_chunk + select_1, postfix, history, histpos)

def peco_history(doc, prefix, postfix, history, histpos):
    return fuzzy_filter_history(doc, prefix, postfix, history, histpos, 'peco')

def fzf_history(doc, prefix, postfix, history, histpos):
    return fuzzy_filter_history(doc, prefix, postfix, history, histpos, 'fzf')


# change the table below if you like, but don't forget to bind the corresponding keys  to 'rlwrap-hotkey' in your .inputrc, or the hotkeys won't work!

keymap = {
    "y" : yank_clipboard,
    "n" : edit_history,
    "p" : peco_history,
    "f" : fzf_history,
    "r" : peco_history,
    "t" : date_in_echo_area
}

############################## helper functions #########################################################

def qw(s):
#    return tuple(s.split())
    return s.split()


def document_all_hotkeys():
    doclist = ''
    dontcare = (None, None, None, None) # dummy arguments for getting the docstring
    for k in 'abcdefghijklmnopqrstuvwxyz':
        try:
            handler = keymap[k]
            if (handler):
                doclist = doclist + "CTRL+{0}:   ".format(k) + handler(1, *dontcare) + "\n"
        except:
            
            pass
    inputrc = "{0}/.inputrc".format(os.environ['HOME'])
    doclist = doclist + "Currently bound hotkeys in .inputrc:\n"
    doclist = doclist + safe_backtick(["grep", "rlwrap-hotkey", inputrc])
    return doclist


def safe_backtick(command_args):
    with subprocess.Popen(command_args, stdout=subprocess.PIPE, universal_newlines=True) as p:
        #result = map(lambda b: b.decode("utf-8"), p.stdout.readlines())
        (result, error) = p.communicate()
    return ''.join(result).rstrip()


# give back corresponding CTRL-key. E.g: control("m") = "\0x13"
def uncontrol(key):
    return chr(ord(key) + 64).lower()


############################################ The Filter #################################################

if __name__ == '__main__':
    rlwrap_filter = rlwrapfilter.RlwrapFilter()
    name = rlwrap_filter.name
    rlwrap_filter.help_text = '\n'.join([
        "Usage: rlwrap -z {0} <command>\n".format(name),
        "handle hotkeys (but only if bound to 'rlwrap-hotkey' in your .inputrc):\n",
        document_all_hotkeys()
    ])

    rlwrap_filter.hotkey_handler = hotkey

    rlwrap_filter.run()





#!/usr/bin/env python3

"""a demo for doing nothing"""

import sys
import os

if 'RLWRAP_FILTERDIR' in os.environ:
    sys.path.append(os.environ['RLWRAP_FILTERDIR'])
else:
    sys.path.append('.')

import rlwrapfilter

filter = rlwrapfilter.RlwrapFilter()

filter.help_text = "Usage: rlwrap [-options] -z null.py <command>\n"\
                   + "a filter that does nothing"

filter.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Python 3 library for [rlwrap](https://github.com/hanslub42/rlwrap) filters

* Synopsis

    filter = rlwrapfilter.RlwrapFilter(message_handler=do_something)
    filter.help_text = 'useful help'
    filter.output_handler = lambda x: re.sub('apple', 'orange', x) # re−write output
    filter.prompt_handler = munge_prompt
    filter.completion_handler = complete_handler
    filter.history_handler = lambda x: re.sub(r'(identified\\s+by\\s+)(\\S+)', r'\\1xXxXxXxX', x)
    filter.run()

This is an [RlwrapFilter](https://github.com/hanslub42/rlwrap/wiki/RlwrapFilter.pm-manpage)
clone written in Python 3. The focus is on providing the same API's
and usage of the Perl version of [RlwrapFilter](https://github.com/hanslub42/rlwrap/wiki/RlwrapFilter.pm-manpage)
as possible.

[rlwrap](https://github.com/hanslub42/rlwrap) is a tiny
utility that sits between the user and any console command, in order
to bestow readline capabilities (line editing, history recall) to
commands that don't have them.

Since version 0.32, rlwrap can use filters to script almost every
aspect of rlwrap's interaction with the user: changing the history,
re-writing output and input, calling a pager or computing completion
word lists from the current input.

rlwrapfilter.py makes it very simple to write rlwrap
filters in Python 3. A filter only needs to instantiate a RlwrapFilter
object, change a few of its default handlers and then call its 'run'
method.

This module and numerous python filters were contributed by Hisanobu Okuda.
"""


import sys
import os
import io
import types
import time
import struct
import select
import re
import traceback
import binascii
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable
import numbers

TAG_INPUT                       = 0
TAG_OUTPUT                      = 1
TAG_HISTORY                     = 2
TAG_COMPLETION                  = 3
TAG_PROMPT                      = 4
TAG_HOTKEY                      = 5
TAG_SIGNAL                      = 6
TAG_WHAT_ARE_YOUR_INTERESTS     = 127
TAG_IGNORE                      = 251
TAG_ADD_TO_COMPLETION_LIST      = 252
TAG_REMOVE_FROM_COMPLETION_LIST = 253
TAG_OUTPUT_OUT_OF_BAND          = 254
TAG_ERROR                       = 255


REJECT_PROMPT = '_THIS_CANNOT_BE_A_PROMPT_'


# we want to behave differently when running outside rlwrap
we_are_running_under_rlwrap = 'RLWRAP_COMMAND_PID' in os.environ

# rlwrap version
rlwrap_version_string  = os.environ.get('RLWRAP_VERSION', "0.41") # e.g. 0.45.2b
major, rest            = rlwrap_version_string.split('.', 1)      # 0, 45.2b
rest                   = re.sub(r'[^\d]','', rest)                # 452
rlwrap_version         = float(major + '.' + rest)                # 0.452


# open communication lines with rlwrap (or with the terminal when not running under rlwrap)
if (we_are_running_under_rlwrap):
    CMD_IN = int(os.environ['RLWRAP_MASTER_PTY_FD'])
    CMD_OUT = int(os.environ['RLWRAP_MASTER_PTY_FD'])

    FILTER_IN = int(os.environ['RLWRAP_INPUT_PIPE_FD'])
    FILTER_OUT = int(os.environ['RLWRAP_OUTPUT_PIPE_FD'])
else:
    CMD_IN = sys.stdout.fileno()
    CMD_OUT = sys.stdin.fileno()

    FILTER_IN = sys.stdout.fileno()
    FILTER_OUT = sys.stdin.fileno()



def when_defined(maybe_ref_to_sub, *args):
    """
`    when_defined(f, x, y, ...) returns f(x, y, ...) if f is defined, x otherwise
    """
    if (maybe_ref_to_sub is not None):
        try:
            return maybe_ref_to_sub(*args)
        except Exception as e:
            send_error(
                "improper handler <{0}> of type {1} (expected a ref to a sub)\n{2}"
                .format(maybe_ref_to_sub, type(maybe_ref_to_sub), traceback.format_exc()),e)
    else:
        return args[0]


def out_of_band(tag):
    return tag > 128


def read_until(fh, stoptext, timeout,
               prompt_search_from=0, prompt_search_to=None):
    """
    read chunks from pty pointed to by fh until either inactive for timeout or stoptext is seen at end-of-chunk
    """
    res = ''
    while (True):
        chunk = read_chunk(fh, timeout);
        if(not chunk):
            # got "" back: timeout
            #send_warn("read_until: timeout")
            return res
        res = res + chunk
        # multi-line mode so that "^" matches a head of each line
        slice = res[prompt_search_from:prompt_search_to]
        if re.search(stoptext, slice, re.MULTILINE):
            return res


def read_chunk(fh, timeout):
    """
    read chunk from pty pointed to by fh with timeout if timed out, returns 0-length string
    """
    if (len(select.select([fh], [], [], timeout)[0]) > 0):
        chunk = os.read(fh, 2**16); # read up-to 2^16=65536 bytes
        return chunk.decode(sys.stdin.encoding, errors="ignore")
    return ""


def read_patiently(fh, count):
    """
    keep reading until count total bytes were read from filehandle fh
    """
    already_read = 0
    buf = bytearray()
    while(already_read < count):
        buf += os.read(fh, count-already_read)
        nread = len(buf)
        if (nread == 0):
            break
        already_read += nread
    return buf


def write_patiently(fh, buffer):
    """
    keep writing until all bytes from $buffer were written to $fh
    """
    already_written = 0
    count = len(buffer)
    while(already_written < count):
        try:
            nwritten = os.write(fh, buffer[already_written:])
            if (nwritten <= 0):
                send_error("error writing: {0}".format(str(buffer)))
            already_written = already_written + nwritten
        except BrokenPipeError: # quit when rlwrap dies
            sys.exit(1)

def read_message():
    """
    read message (tag, length word and contents) from FILTER_IN
    """
    if not we_are_running_under_rlwrap:
        return read_from_stdin()

    tag = int.from_bytes(read_patiently(FILTER_IN,1), sys.byteorder)
    length = int.from_bytes(read_patiently(FILTER_IN,4), sys.byteorder)
    message = read_patiently(FILTER_IN, length).decode(sys.stdin.encoding, errors = "ignore")
    # \Z matches only at the end of the string in python
    message = re.sub(r'\n\Z', '', str(message or ""))
    return tag, message


def write_message(tag, message):
    if (not we_are_running_under_rlwrap):
        return write_to_stdout(tag, message)

    message = '\n' if message is None else message + '\n'  # allow undefined message
    bmessage = bytearray(message, sys.stdin.encoding)
    length = len(bmessage)

    write_patiently(FILTER_OUT, tag.to_bytes(1, sys.byteorder, signed=False))
    write_patiently(FILTER_OUT, length.to_bytes(4, sys.byteorder, signed=False))
    write_patiently(FILTER_OUT, bmessage)


def read_from_stdin():
    tag = None
    prompt = None
    tagname = None
    while (tag is None):
        try:
            m = re.match(r'(\S+) (.*?)\r?\n', sys.stdin.readline())
        except KeyboardInterrupt:
            sys.exit()
        if not m:
            sys.exit()
        tagname, message = m.groups()
        tagname=tagname.upper()            # allow lowercase tag names
        if not re.match(r'TAG_', tagname): # allow omitting the tiresome "TAG_" prefix
            tagname = "TAG_" + tagname
        message.replace("\\t","\t").replace("\\n","\n")
        tag = name2tag(tagname)
    return tag, message

def name2tag(name):
        """
        Convert a valid tag name like " TAG_PROMPT " to a tag (an integer)
        """
        try:
            tag = eval(name)
        except Exception as e:
            raise SystemExit('unknown tagname {0}'.format(name))
        return tag


def tag2name(tag):
    """
    Convert the tag (an integer) to its name (e.g. " TAG_PROMPT ")
    """
    for name in ['TAG_REMOVE_FROM_COMPLETION_LIST',
                 'TAG_ADD_TO_COMPLETION_LIST',
                 'TAG_WHAT_ARE_YOUR_INTERESTS',
                 'TAG_INPUT',
                 'TAG_PROMPT',
                 'TAG_COMPLETION',
                 'TAG_HOTKEY',
                 'TAG_SIGNAL',
                 'TAG_HISTORY',
                 'TAG_OUTPUT_OUT_OF_BAND',
                 'TAG_ERROR',
                 'TAG_IGNORE',
                 'TAG_OUTPUT']:
        if (eval('{0} == {1}'.format(str(tag), name))):
            return name




def write_to_stdout(tag, message):
    print('{0}: {1}'.format(tag2name(tag), message))


def send_warn(message):
    """
    send message to rlwrap.
    """
    write_message(TAG_OUTPUT_OUT_OF_BAND, "{0}: {1}".format(__name__, message))


def send_error(message,e = None):
    """
    send message to rlwrap, and raise e.
    """
    write_message(TAG_OUTPUT_OUT_OF_BAND if e else TAG_ERROR, "{0}: {1}".format(__name__, message))
    if e:
       raise e
    else:
       time.sleep(2) # timeout doesn't matter much because rlwrap will kill us anyway
       exit()

def intercept_error(func):
    """
    A decorator to intercept an exception, send the message to rlwrap, and raise an exception.
    """
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            write_message(
                TAG_ERROR,
                "{0}: {1}".format(__name__, '/'.join(map(str,e.args)))
            )
            raise e
    return wrapper



def intercept_error_with_message(message=None):
    """  
    A decorator (-factory) to intercept an exception, send the message
    to rlwrap, print a message and exit (or re-raise the exception, if
    message = None) N.B: decorators, hence also <message> are evaluated
    at compile time. @intercept_error_with_message (f"This script
    crashed after {sec} seconds") doesn't make sense.
    """
    def intercept_error_closure(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                complete_message = "{0}: {1}".format(__name__, '/'.join(map(str,e.args))) if message == None else "{0}\n{1}".format(message, e)
                write_message(TAG_ERROR, complete_message)
                if message:
                    exit()
                else:
                    raise e
        return wrapper
    return intercept_error_closure

def is_string(value):
    return isinstance(value, str) or value == None


def is_boolean(value):
    return isinstance(value, bool) or value == None


def is_integer(value):
    return isinstance(value, int) or value == None


def is_float(value):
    return isinstance(value, numbers.Number) or value == None


def is_callable(value):
    return isinstance(value, Callable) or value == None


@intercept_error
def test_intercept():
    print('test intercept!!! + + + +')
    raise Exception('test exception = = = = = .........')


DIGIT_NUMBER=8

def split_rlwrap_message(message):
    bmessage = bytes(message, sys.stdin.encoding)
    fields = []

    while(len(bmessage) != 0):
        blen = bmessage[:DIGIT_NUMBER]
        bmessage = bmessage[DIGIT_NUMBER:]
        length = int(str(blen, sys.stdin.encoding), base=16)
        bfield = bmessage[:length]
        bmessage = bmessage[length:]
        fields.append(str(bfield, sys.stdin.encoding))
    return fields


def merge_fields(fields):
    message = ""

    for field in fields:
        length = len(bytes(field, sys.stdin.encoding))
        lenstr = format(length, '0' + str(DIGIT_NUMBER) + 'x')
        message = message + lenstr + field
    return message


class RlwrapFilterError(Exception):
    """
    A custom exception for rlwrap
    """
    def __init__(self, *args):
        super().__init__(args)


class RlwrapFilter:
    """
    class for rlwrap filters
    """

    def __setattr__(self, name, value):
        if not name in self._fields:
            self.warn("There is no '{0}' attribute in class {1}\n".format(name, self.__class__.__name__))

        is_valid_type = self._field_types[name]
        if not is_valid_type(value):
            self.warn("{0} should not be '{1}'\n".format(name, type(value)))

        if name == 'minimal_rlwrap_version' and (value > rlwrap_version):
            self.error("requires rlwrap version {0} or newer.\n".format(str(value)))
        self.__dict__[name] = value


    """
    def __getattr__(self, name):
        if(name in self.fields):
            return self.__dict__[name]
        elif(name in self.handlers):
            return self.__dict__[name]
        else:
            send_error("There is no '{0}' attribute in class {1}"
                       .format(name, self.__class__.__name__))
    """


    def __init__(self, **kwargs):
        self.__dict__['_field_types'] = {
            'input_handler':is_callable,
            'output_handler':is_callable,
            'prompt_handler':is_callable,
            'hotkey_handler':is_callable,
            'signal_handler':is_callable,
            'echo_handler':is_callable,
            'message_handler':is_callable,
            'history_handler':is_callable,
            'completion_handler':is_callable,
            'help_text':is_string,
            'cloak_and_dagger_verbose':is_boolean,
            'cumulative_output':is_string,
            'prompts_are_never_empty':is_boolean,
            'previous_tag':is_integer,
            'previous_message':is_string,
            'echo_has_been_handled':is_boolean,
            'saved_output':is_string,
            'prompt_rejected':is_string,
            'command_line':is_string,
            'running_under_rlwrap':is_boolean,
            'minimal_rlwrap_version':is_float,
            'name':is_string
        }

        self.__dict__['_fields'] = self.__dict__['_field_types'].keys()

        for field in self._fields:
            self.__dict__[field] = None

        self.previous_tag = -1
        self.echo_has_been_handled = False
        self.saved_output = ''
        self.cumulative_output = ''
        self.minimal_rlwrap_version = rlwrap_version
        self.command_line = os.environ.get('RLWRAP_COMMAND_LINE')
        self.running_under_rlwrap = 'RLWRAP_COMMAND_PID' in os.environ
        self.name = os.path.basename(sys.argv[0])

        for key in kwargs:
            exec('self.{0} = kwargs[key]'.format(key))



    def handle_output(self, message):
        """
        split output in echo and the rest and call the appropriate handlers on them
        """
        (handled_echo, nl) = ('', '')
        if (self.previous_tag is not None and self.previous_tag == TAG_INPUT):
            self.cumulative_output = ""
            self.echo_has_been_handled = False

        if (not self.echo_has_been_handled):
            if (not re.search(r'\n', message)):
                # save all output until we have one *whole* echo line
                self.saved_output = self.saved_output + message
                return ""
            else:
                # ... then process it
                message = self.saved_output + message
                self.echo_has_been_handled = True
                self.saved_output = ""
                (echo, nl, message) = re.match(r'^([^\n\r]*)(\r?\n)?(.*)?', message, re.DOTALL).groups()
                handled_echo = when_defined(self.echo_handler, echo)
        self.cumulative_output = self.cumulative_output + message
        return handled_echo + str(nl or "") + str(when_defined(self.output_handler, message))


    def add_to_completion_list(self, *args):
        write_message(TAG_ADD_TO_COMPLETION_LIST, ' '.join(args))


    def remove_from_completion_list(self, *args):
        write_message(TAG_REMOVE_FROM_COMPLETION_LIST, ' '.join(args))


    def cloak_and_dagger(self, question, prompt, timeout,
                         prompt_search_from=0, prompt_search_to=None):
        """
        have a private chat with the rlwrapped command. This relies very much om the assumption that command stops. 
        talking, and only listens, when it has displayed the prompt
        """
        write_patiently(CMD_IN, bytearray(question + "\n", sys.stdin.encoding))
        if (self.cloak_and_dagger_verbose):
            self.send_output_oob("cloak_and_dagger question: {0}\n".format(question))
        response = read_until(CMD_OUT, prompt, timeout,
                              prompt_search_from=prompt_search_from,
                              prompt_search_to=prompt_search_to)
        response = re.sub(r'^.*?\n', '', response) # chop off echoed question;
        response = re.sub(r'{0}$'.format(prompt), '', response) # chop off prompt;
        if (self.cloak_and_dagger_verbose):
            self.send_output_oob("cloak_and_dagger response: {0}\n".format(response))
        return response


    def vacuum_stale_message(self, prompt, timeout):
        """
        Some command returns messages asynchronously
        and tends to delay message when invoking multiple `cloak_and_dagger`.
        You may want to drop message at such time.

        rlwrap_filter.cloak_and_dagger(command1, prompt, timeout)
        rlwrap_filter.cloak_and_dagger(command2, prompt, timeout)
        ...
        time.sleep(1)
        rlwrap_filter.vacuum_stale_message(prompt, timeout)
        """
        response = read_until(CMD_OUT, prompt, timeout)
        return response


    def add_interests(self, message):
        interested = list(message)
        tag2handler = {TAG_OUTPUT      : self.output_handler or self.echo_handler, # echo is the first OUTPUT after an INPUT
                       TAG_INPUT       : self.input_handler or self.echo_handler,  # so to determine what is ECHO we need to see INPUT... 
                       TAG_HISTORY     : self.history_handler,
                       TAG_COMPLETION  : self.completion_handler,
                       TAG_PROMPT      : self.prompt_handler,
                       TAG_HOTKEY      : self.hotkey_handler,
                       TAG_SIGNAL      : self.signal_handler}

        for tag in range(0, len(message)):
            if interested[tag] == 'y':
                continue   # a preceding filter in the pipeline has already shown interest
            if tag2handler[tag] is not None:
                interested[tag] = 'y'
        return ''.join(interested)
    
    def name2tag(self, name):
        """
        Convert a valid tag name like " TAG_PROMPT " to a tag (an integer)
        """
        return name2tag(name)

    def tag2name(self, tag):
        """
        Convert the tag (an integer) to its name (e.g. " TAG_PROMPT ")
        """
        return tag2name(tag)

    def warn(self, message):
        """
        send message to rlwrap.
        """
        send_warn(message)


    def error(self, message,e = None):
        """
        send message to rlwrap, and raise e.
        """
        send_error(message, e)


    def send_output_oob(self, text):
        write_message(TAG_OUTPUT_OUT_OF_BAND, text)


    def send_ignore_oob(self, text):
        write_message(TAG_IGNORE, text)

      
    def tweak_readline_oob(self, rl_tweak, *args):
        nargs = {'rl_variable_bind'                   : 2,
                 'rl_completer_word_break_characters' : 1,
                 'rl_completer_quote_characters'      : 1,
                 'rl_filename_completion_desired'     : 1}
                 # the above list can be extended in future versions
        if rl_tweak not in nargs: 
            self.error("tweak_readline_oob() called with unknown/unimplemented readline function '{}'".format(rl_tweak))
        if len(args) !=  nargs[rl_tweak]:
            self.error("tweak_readline_oob({},...) should be called with exactly {} args".format(rl_tweak, nargs[rl_tweak] + 1))
        self.send_ignore_oob("@" + "::".join((rl_tweak,) + args + ("\n",)))

        
    def cwd(self):
        return os.getcwd()

    def run(self):
        """
        event loop
        """

        # $RLWRAP_COMMAND_PID can be undefined (e.g. when run interactively, or under rlwrap -z listing
        # or == "0" (when rlwrap is called without a command name, like in rlwrap -z filter.py)
        # In both cases: print help text
        if os.environ.get('RLWRAP_COMMAND_PID') in [None, '0']:
            write_message(TAG_OUTPUT_OUT_OF_BAND, self.help_text + '\n')
        while(True):
            tag, message = read_message()
            
            message = when_defined(self.message_handler, message, tag) # ignore return value

            if (tag == TAG_INPUT):
                response = when_defined(self.input_handler, message)
            elif (tag == TAG_OUTPUT):
                response = self.handle_output(message)
            elif (tag == TAG_HISTORY):
                response = when_defined(self.history_handler, message)
            elif (tag == TAG_COMPLETION):
                if (self.completion_handler is not None):
                    params = split_rlwrap_message(message)
                    (line, prefix, completions) = (params[0], params[1], params[2:])
                    completions = self.completion_handler(line, prefix, completions)
                    response = merge_fields([line, prefix] + completions)
                else:
                    response = message
            elif (tag == TAG_HOTKEY):
                if (self.hotkey_handler is not None):
                    params = split_rlwrap_message(message)
                    result = self.hotkey_handler(*params)
                    response = merge_fields(result)
                else:
                    response = message
            elif (tag == TAG_PROMPT):
                if (message == REJECT_PROMPT or
                    (self.prompts_are_never_empty is not None and message == '')):
                    write_message(tag,REJECT_PROMPT);
                    # don't update <previous_tag> and don't reset <cumulative_input>
                    next
                if (os.environ.get('RLWRAP_IMPATIENT') and not re.search(r'\n$', self.cumulative_output)):
                    # cumulative output contains prompt: chop it off!
                    # s/[^\n]*$// takes way too long on big strings,
                    # what is the optimal regex to do this?
                    self.cumulative_output = re.sub(r'(?<![^\n])[^\n]*$', '', self.cumulative_output)

                response = when_defined(self.prompt_handler, message)
                if (re.search(r'\n', response)):
                    send_error('prompts may not contain newlines!')
            elif (tag == TAG_SIGNAL):
                response = when_defined(self.signal_handler, message)
            elif (tag == TAG_WHAT_ARE_YOUR_INTERESTS):
                response = self.add_interests(message)
            else:
                # No error message, compatible with future rlwrap
                # versions that may define new tag types
                response = message

            if (not (out_of_band(tag) and (tag == TAG_PROMPT and response == REJECT_PROMPT))):
                self.previous_tag = tag
                self.previous_message = message

            write_message(tag, response)




if __name__ == '__main__':
    import doctest
    doctest.testmod()
